@@ -8,6 +8,7 @@
  * Contributors:
  *     IBM Corporation - initial API and implementation
  *     Markus Schorn (Wind River) - [108066] Project prefs marked dirty on read
+ *     James Blackburn (Broadcom Corp.) - ongoing development
  *******************************************************************************/
 package org.eclipse.core.internal.resources;
 
@@ -36,47 +37,41 @@ public class ProjectPreferences extends EclipsePreferences {
 
 	class SortedProperties extends Properties {
 
-		class IteratorWrapper implements Enumeration {
-			Iterator iterator;
-
-			public IteratorWrapper(Iterator iterator) {
-				this.iterator = iterator;
-			}
-
-			public boolean hasMoreElements() {
-				return iterator.hasNext();
-			}
-
-			public Object nextElement() {
-				return iterator.next();
-			}
-		}
-
 		private static final long serialVersionUID = 1L;
 
 		/* (non-Javadoc)
 		 * @see java.util.Hashtable#keys()
 		 */
-		public synchronized Enumeration keys() {
-			TreeSet set = new TreeSet();
-			for (Enumeration e = super.keys(); e.hasMoreElements();)
+		public synchronized Enumeration<Object> keys() {
+			TreeSet<Object> set = new TreeSet<Object>();
+			for (Enumeration<Object> e = super.keys(); e.hasMoreElements();)
 				set.add(e.nextElement());
-			return new IteratorWrapper(set.iterator());
+			return Collections.enumeration(set);
+		}
+
+		/* (non-Javadoc)
+		 * @see java.util.Hashtable#entrySet()
+		 */
+		public Set<Map.Entry<Object, Object>> entrySet() {
+			TreeSet<Map.Entry<Object, Object>> set = new TreeSet<Map.Entry<Object, Object>>(new Comparator<Map.Entry<Object, Object>>() {
+				public int compare(Map.Entry<Object, Object> e1, Map.Entry<Object, Object> e2) {
+					String s1 = (String) e1.getKey();
+					String s2 = (String) e2.getKey();
+					return s1.compareTo(s2);
+				}
+			});
+			for (Iterator<Map.Entry<Object, Object>> i = super.entrySet().iterator(); i.hasNext();)
+				set.add(i.next());
+			return set;
 		}
 	}
-	
-	/**
-	 * Name of a preference for configuring whether encodings for derived resources 
-	 * within the project should be stored in a separate derived preference file.
-	 */
-	public static final String PREF_SEPARATE_DERIVED_ENCODINGS = "separateDerivedEncodings"; //$NON-NLS-1$
 
 	static final String PREFS_REGULAR_QUALIFIER = ResourcesPlugin.PI_RESOURCES;
 	static final String PREFS_DERIVED_QUALIFIER = PREFS_REGULAR_QUALIFIER + ".derived"; //$NON-NLS-1$
 	/**
 	 * Cache which nodes have been loaded from disk
 	 */
-	protected static Set loadedNodes = Collections.synchronizedSet(new HashSet());
+	protected static Set<String> loadedNodes = Collections.synchronizedSet(new HashSet<String>());
 	private IFile file;
 	private boolean initialized = false;
 	/**
@@ -280,8 +275,8 @@ public class ProjectPreferences extends EclipsePreferences {
 	private static void removeLoadedNodes(Preferences node) {
 		String path = node.absolutePath();
 		synchronized (loadedNodes) {
-			for (Iterator i = loadedNodes.iterator(); i.hasNext();) {
-				String key = (String) i.next();
+			for (Iterator<String> i = loadedNodes.iterator(); i.hasNext();) {
+				String key = i.next();
 				if (key.startsWith(path))
 					i.remove();
 			}
@@ -349,21 +344,7 @@ public class ProjectPreferences extends EclipsePreferences {
 		if (segmentCount > 2)
 			qualifier = getSegment(path, 2);
 
-		if (segmentCount != 2)
-			return;
-
-		// else segmentCount == 2 so we initialize the children
-		if (initialized)
-			return;
-		try {
-			synchronized (this) {
-				String[] names = computeChildren();
-				for (int i = 0; i < names.length; i++)
-					addChild(names[i], null);
-			}
-		} finally {
-			initialized = true;
-		}
+		initialize();
 	}
 
 	/*
@@ -382,13 +363,13 @@ public class ProjectPreferences extends EclipsePreferences {
 		} catch (CoreException e) {
 			return EMPTY_STRING_ARRAY;
 		}
-		ArrayList result = new ArrayList();
+		ArrayList<String> result = new ArrayList<String>();
 		for (int i = 0; i < members.length; i++) {
 			IResource resource = members[i];
 			if (resource.getType() == IResource.FILE && PREFS_FILE_EXTENSION.equals(resource.getFullPath().getFileExtension()))
 				result.add(resource.getFullPath().removeFileExtension().lastSegment());
 		}
-		return (String[]) result.toArray(EMPTY_STRING_ARRAY);
+		return result.toArray(EMPTY_STRING_ARRAY);
 	}
 
 	public void flush() throws BackingStoreException {
@@ -399,30 +380,6 @@ public class ProjectPreferences extends EclipsePreferences {
 			super.flush();
 		} finally {
 			isWriting = false;
-			if ((segmentCount == 3) && (PREFS_DERIVED_QUALIFIER.equals(qualifier))) {
-				final IFile fileInWorkspace = getFile();
-				if (fileInWorkspace != null) {
-					IWorkspaceRunnable operation = new IWorkspaceRunnable() {
-						public void run(IProgressMonitor monitor) throws CoreException {
-							if (fileInWorkspace.exists())
-								fileInWorkspace.setDerived(true, monitor);
-						}
-					};
-					Workspace workspace = ((Workspace) ResourcesPlugin.getWorkspace());
-					try {
-						if (workspace.getWorkManager().isLockAlreadyAcquired())
-							operation.run(null);
-						else
-							workspace.run(operation, workspace.getRuleFactory().derivedRule(fileInWorkspace), IResource.NONE, null);
-					} catch (OperationCanceledException e) {
-						throw new BackingStoreException(Messages.preferences_operationCanceled);
-					} catch (CoreException e) {
-						String message = NLS.bind(Messages.preferences_setDerivedException, fileInWorkspace.getFullPath());
-						log(new Status(IStatus.ERROR, ResourcesPlugin.PI_RESOURCES, IStatus.ERROR, message, e));
-						throw new BackingStoreException(message);
-					}
-				}
-			}
 		}
 	}
 
@@ -474,9 +431,9 @@ public class ProjectPreferences extends EclipsePreferences {
 
 	protected String internalPut(String key, String newValue) {
 		if ((segmentCount == 3) && PREFS_REGULAR_QUALIFIER.equals(qualifier) && (project != null)) {
-			if (ProjectPreferences.PREF_SEPARATE_DERIVED_ENCODINGS.equals(key)) {
+			if (ResourcesPlugin.PREF_SEPARATE_DERIVED_ENCODINGS.equals(key)) {
 				Workspace workspace = ((Workspace) ResourcesPlugin.getWorkspace());
-				if (Boolean.valueOf(newValue).booleanValue())
+				if (Boolean.parseBoolean(newValue))
 					workspace.getCharsetManager().splitEncodingPreferences(project);
 				else
 					workspace.getCharsetManager().mergeEncodingPreferences(project);
@@ -485,6 +442,29 @@ public class ProjectPreferences extends EclipsePreferences {
 		return super.internalPut(key, newValue);
 	}
 
+	protected void initialize() {
+		if (segmentCount != 2)
+			return;
+
+		// if already initialized, then skip this initialization
+		if (initialized)
+			return;
+
+		// initialize the children only if project is opened
+		if (project.isOpen()) {
+			try {
+				synchronized (this) {
+					String[] names = computeChildren();
+					for (int i = 0; i < names.length; i++)
+						addChild(names[i], null);
+				}
+			} finally {
+				// mark as initialized so that subsequent project opening will not initialize preferences again
+				initialized = true;
+			}
+		}
+	}
+
 	protected boolean isAlreadyLoaded(IEclipsePreferences node) {
 		return loadedNodes.contains(node.absolutePath());
 	}
@@ -500,6 +480,9 @@ public class ProjectPreferences extends EclipsePreferences {
 				Policy.debug("Unable to determine preference file or file does not exist for node: " + absolutePath()); //$NON-NLS-1$
 			return;
 		}
+		// Before loading the preferences, try to initialize parent node (/project/<projectName>)
+		// see bug 335591
+		((ProjectPreferences) parent).initialize();
 		if (Policy.DEBUG_PREFERENCES)
 			Policy.debug("Loading preferences from file: " + localFile.getFullPath()); //$NON-NLS-1$
 		Properties fromDisk = new Properties();
@@ -519,9 +502,6 @@ public class ProjectPreferences extends EclipsePreferences {
 			FileUtil.safeClose(input);
 		}
 		convertFromProperties(this, fromDisk, true);
-	}
-
-	protected void loaded() {
 		loadedNodes.add(absolutePath());
 	}
 
@@ -549,7 +529,7 @@ public class ProjectPreferences extends EclipsePreferences {
 	public void remove(String key) {
 		super.remove(key);
 		if ((segmentCount == 3) && PREFS_REGULAR_QUALIFIER.equals(qualifier) && (project != null)) {
-			if (ProjectPreferences.PREF_SEPARATE_DERIVED_ENCODINGS.equals(key)) {
+			if (ResourcesPlugin.PREF_SEPARATE_DERIVED_ENCODINGS.equals(key)) {
 				Workspace workspace = ((Workspace) ResourcesPlugin.getWorkspace());
 				workspace.getCharsetManager().mergeEncodingPreferences(project);
 			}
@@ -600,18 +580,16 @@ public class ProjectPreferences extends EclipsePreferences {
 			ByteArrayOutputStream output = new ByteArrayOutputStream();
 			try {
 				table.store(output, null);
+				output.close();
 			} catch (IOException e) {
 				String message = NLS.bind(Messages.preferences_saveProblems, absolutePath());
 				log(new Status(IStatus.ERROR, Platform.PI_RUNTIME, IStatus.ERROR, message, e));
 				throw new BackingStoreException(message);
 			} finally {
-				try {
-					output.close();
-				} catch (IOException e) {
-					// ignore
-				}
+				FileUtil.safeClose(output);
 			}
 			final InputStream input = new BufferedInputStream(new ByteArrayInputStream(output.toByteArray()));
+			final String finalQualifier = qualifier;
 			IWorkspaceRunnable operation = new IWorkspaceRunnable() {
 				public void run(IProgressMonitor monitor) throws CoreException {
 					if (fileInWorkspace.exists()) {
@@ -636,6 +614,8 @@ public class ProjectPreferences extends EclipsePreferences {
 							Policy.debug("Creating preference file: " + fileInWorkspace.getLocation()); //$NON-NLS-1$
 						fileInWorkspace.create(input, IResource.NONE, null);
 					}
+					if (PREFS_DERIVED_QUALIFIER.equals(finalQualifier))
+						fileInWorkspace.setDerived(true, null);
 				}
 			};
 			//don't bother with scheduling rules if we are already inside an operation
@@ -643,8 +623,8 @@ public class ProjectPreferences extends EclipsePreferences {
 				if (((Workspace) workspace).getWorkManager().isLockAlreadyAcquired()) {
 					operation.run(null);
 				} else {
-					// we might: create the .settings folder, create the file, or modify the file.
-					ISchedulingRule rule = MultiRule.combine(factory.createRule(fileInWorkspace.getParent()), factory.modifyRule(fileInWorkspace));
+					// we might: create the .settings folder, create the file, modify the file, or set derived flag for the file.
+					ISchedulingRule rule = MultiRule.combine(new ISchedulingRule[] {factory.createRule(fileInWorkspace.getParent()), factory.modifyRule(fileInWorkspace), factory.derivedRule(fileInWorkspace)});
 					workspace.run(operation, rule, IResource.NONE, null);
 				}
 			} catch (OperationCanceledException e) {
